import aiohttp
import asyncio
import json
import time
import signal
from typing import Optional, List

from .exceptions import EventTypeError
from .user import User, UserConnection
from .guild import Guild, GuildMember
from .base import DiscordObject
from .constants import DISCORD_API_URL
from .channel import Channel
from .http import HTTPHandler
from .websocket import DiscordWebsocket

import logging
logger = logging.getLogger(__name__)

# TODO: Delete this when sure.
# def task_handler() -> None:
#     logger.debug('Stopping tasks')
#     for task in asyncio.Task.all_tasks():
#         task.cancel()


class DiscordBot:
    """This class represents a discord bot object, it has many utility methods for making a bot.

    .. versionadded:: 0.2.0

    Attributes:
        token (:obj:`str`): The discord token used for authentication
        http (:class:`.HTTPHandler`): Used for making http requests and websocket creation.
        guilds (:obj:`list` of :class:`.Guild`): The list of guilds the bot is in.
        user (:class:`.User`): The user object of the bot.
        ws (:class:`.DiscordWebsocket`): The websocket used for communication
    """

    def __init__(self, token: str):
        """DiscordBot constructor.

        Args:
            token (obj:`str`): The discord token used for authentication.
        """
        self.token: str = token
        self.http: HTTPHandler = HTTPHandler(token, self)
        self.guilds: List[Guild] = []
        self.loop = asyncio.get_event_loop()
        self.do_sync = self.loop.run_until_complete
        self.user: Optional[User] = None
        self.ws: Optional[DiscordWebsocket] = None

    def run(self) -> None:
        """Starts the bot, making it connect to discord.

        .. versionadded:: 0.2.0
        """
        try:
            self.do_sync(self._start())
        except KeyboardInterrupt:
            pass
        except asyncio.CancelledError:
            logger.debug('Tasks has been cancelled')
        finally:
            self.do_sync(self.http.close_session())
            self.loop.close()

    async def raise_event(self, event: str, *args, **kwargs) -> None:
        try:
            if hasattr(self, event):
                await getattr(self, event)(*args, **kwargs)
        except asyncio.CancelledError as e:
            logger.error(e, exc_info=1)
            pass
        except Exception as e:
            logger.error(e, exc_info=1)
            # TODO: handle error here
            pass

    def event(self, name: str=None):
        """DiscordBot event decorator, uses the function's name or the 'name' parameter to subscribe to a event

        .. versionadded:: 0.2.0
        """
        def real_event(coro):
            if not asyncio.iscoroutinefunction(coro):
                raise EventTypeError(
                    'The event function must be a coroutine.')
            if name is not None:
                if not isinstance(name, str):
                    raise TypeError('event name must be of type str')
                else:
                    try:
                        getattr(self, name)
                        raise AttributeError(
                            'tried to subscribe to a event that doesn\'t exist, or you already subscribed to it.')
                    except AttributeError:
                        pass
                    finally:
                        setattr(self, name, coro)
            else:
                setattr(self, coro.__name__, coro)
            logger.debug(f'{coro.__name__} subscribed succesfully.')
            return coro
        return real_event

    async def _start(self):
        """Used internally by the bot to start the http session and the websocket

        .. versionadded:: 0.2.0
        """
        await self.http.create_session()
        self.ws = DiscordWebsocket(self.http)
        await self.ws.start()

    async def exit(self):
        """Disconnects the bot

        .. versionadded:: 0.2.0
        """
        if not self.ws.closed:
            await self.ws.close()
        await self.http.close_session()

    async def change_avatar(self, url: str):
        raise NotImplementedError()
        # TODO: Implement this: https://discordapp.com/developers/docs/resources/user#modify-current-user

    async def get_user(self, id: int) -> User:
        """Gets the user object from the given user id.

        .. versionadded:: 0.2.0

        Args:
            id (:obj:`int`): The user id

        Returns:
            :class:`.User`: The requested user
        """
        res = await self.http.request_url('/users/' + str(id))
        return await User.from_api_res(res)

    async def get_self_user(self) -> User:
        """Returns the bot user object. (it's like `get_user('@me')`)

        .. versionadded:: 0.2.0
        """
        return await self.get_user("@me")

    async def get_guilds(self) -> list:
        """Returns a list of guilds where the bot is in.

        .. versionadded:: 0.2.0
        """
        res = await self.http.request_url('/users/@me/guilds')
        return await Guild.from_api_res(res)

    async def get_guild(self, guild_id: int) -> Guild:
        """Returns a Guild object from a guild id.

        .. versionadded:: 0.2.0

        Args:
            guild_id (:obj:`int`): The guild id

        Returns:
            :class:`.Guild`: The requested guild
        """
        res = await self.http.request_url('/guilds/' + str(guild_id))
        return await Guild.from_api_res(res)

    async def get_guild_members(self, guild: Guild):
        """Gets and fills the guild with the members info.

        .. versionadded:: 0.2.0

        Args:
            guild (:class:`.Guild`): The guild to fill the members.
        """
        res = await self.http.request_url('/guilds/' + guild.id + '/members')
        await guild._fill_members(res)

    async def get_guild_member(self, guild: Guild, member_id: int) -> GuildMember:
        """Gets a guild member info for the guild and the member id.

        .. versionadded:: 0.2.0

        Args:
            guild (:class:`.Guild`): The guild
            member_id (:obj:`int`): The member id

        Returns:
            :class:`.GuildMember`: The guild member
        """
        res = await self.http.request_url('/guilds/' + guild.id + '/members/' + member_id)
        return await GuildMember.from_api_res(res)

    async def leave_guild(self, guild_id: int):
        """Leaves a guild.

        .. versionadded:: 0.2.0

        Args:
            guild_id (:obj:`int`): The guild id
        """
        await self.http.request_url(f'/users/@me/guilds/{guild_id}', type='DELETE')

    async def get_channel(self, channel_id: int) -> Channel:
        """Gets a channel from it's id

        .. versionadded:: 0.2.0

        Args:
            channel_id (:obj:`int`): The channel id

        Returns:
            :class:`.Channel`: The channel
        """
        res = await self.http.request_url(f'/channels/{channel_id}')
        return await Channel.from_api_res(res)

    async def delete_channel(self, channel_id: int) -> Channel:
        """Deletes a channel.

        Note:
            Delete a channel, or close a private message. Requires the 'MANAGE_CHANNELS' permission for the guild.
            Deleting a category does not delete its child channels; they will have their parent_id removed and a Channel Update Gateway event will fire for each of them.
            Returns a channel object on success.
            Fires a Channel Delete Gateway event.

        .. versionadded:: 0.2.0

        Args:
            channel_id (:obj:`int`): The channel id

        Returns:
            :class:`.Channel`: The deleted channel
        """
        res = await self.http.request_url(f'/channels/{channel_id}', type='DELETE')
        return await Channel.from_api_res(res)

    async def get_dms(self) -> list:
        """Gets a list of dms.

        .. versionadded:: 0.2.0

        Returns:
            :obj:`list` of :class:`.Channel`: The DMs channels
        """
        res = await self.http.request_url('/users/@me/channels')
        return await Channel.from_api_res(res)

    async def get_connections(self) -> list:
        res = await self.http.request_url('/users/@me/connections')
        return await UserConnection.from_api_res(res)


__all__ = [
    'DiscordBot',
]



"""Contains all related Channel Discord objects"""

from .base import DiscordObject
from .user import User
from .role import Role
from .emoji import Emoji


class Overwrite(DiscordObject):
    """Represents a Overwrite object.

    .. versionadded:: 0.2.0

    Attributes:
        id (:obj:`int`): role or user id
        type (:obj:`str`): either "role" or "member"
        allow (:obj:`int`): permission bit set
        deny (:obj:`int`): permission bit set
    """

    def __init__(self, id=0, type="", allow=0, deny=0):
        self.id = id
        self.type = type
        self.allow = allow
        self.deny = deny


class Channel(DiscordObject):
    """Represents a guild or DM channel within Discord.

    .. versionadded:: 0.2.0

    Attributes:
        id (:obj:`int`): the id of this channel
        value_type (:obj:`int`): the value_type of channel
        guild_id (:obj:`int`, optional): the id of the guild
        position (:obj:`int`, optional): sorting position of the channel
        permission_overwrites (:obj:`list` of :class:`.Overwrite`, optional): explicit permission overwrites for members and roles
        name (:obj:`str`, optional): the name of the channel (2-100 characters)
        topic (:obj:`str`, optional): the channel topic (0-1024 characters)
        nsfw (:obj:`bool`, optional): if the channel is nsfw
        last_message_id (:obj:`int`, optional): the id of the last message sent in this channel (may not point to an existing or valid message)
        bitrate (:obj:`int`, optional): the bitrate (in bits) of the voice channel
        user_limit (:obj:`int`, optional): the user limit of the voice channel
        recipients (:obj:`list` of :class:`.User`, optional): the recipients of the DM
        icon (:obj:`str`, optional): icon hash
        owner_id (:obj:`int`, optional): id of the DM creator
        application_id (:obj:`int`, optional): application id of the group DM creator if it is bot-created
        parent_id (:obj:`int`, optional): id of the parent category for a channel
        last_pin_timestamp (:obj:`int`, optional): timestamp when the last pinned message was pinned
    """

    def __init__(self, id=0, type=0, guild_id=0, position=0, permission_overwrites=[],
                 name="", topic="", nsfw=False, last_message_id=0,
                 bitrate=0, user_limit=0, recipients=[], icon="",
                 owner_id=0, application_id=0, parent_id=0, last_pin_timestamp=None):
        self.id = id
        self.type = type
        self.guild_id = guild_id
        self.position = position
        self.permission_overwrites = permission_overwrites
        self.name = name
        self.topic = topic
        self.nsfw = nsfw
        self.last_message_id = last_message_id
        self.bitrate = bitrate
        self.user_limit = user_limit
        self.recipients = recipients
        self.icon = icon
        self.owner_id = owner_id
        self.application_id = application_id
        self.parent_id = parent_id
        self.last_pin_timestamp = last_pin_timestamp

    async def _from_api_ext(self, key, value):
        if key == 'recipients':
            setattr(self, key, [await User.from_api_res(x) for x in value])
        elif key == 'permission_overwrites':
            setattr(self, key, [await Overwrite.from_api_res(x) for x in value])
        else:
            return await super()._from_api_ext(key, value)

    def __str__(self):
        return f'{self.name}#{self.id}'

    def __repr__(self):
        return f'<Channel Object: {self.name}#{self.id}>'


class MessageActivity(DiscordObject):
    """Represents a Message Activity.

    .. versionadded:: 0.2.0

    Attributes:
        type (:obj:`int`): type of message activity
        party_id (:obj:`str`, optional): party_id from a Rich Presence event
    """

    def __init__(self, type=None, party_id=""):
        self.type = type
        self.party_id = party_id


class MessageApplication(DiscordObject):
    """Represents a Message Application.

    .. versionadded:: 0.2.0

    Attributes:
        id (:obj:`int`): id of the application
        cover_image (:obj:`str`): id of the embed's image asset
        description (:obj:`str`): application's description
        icon (:obj:`str`): id of the application's icon
        name (:obj:`str`): name of the application
    """

    def __init__(self, id=0, cover_image="", description="", icon="", name=""):
        self.id = id
        self.cover_image = cover_image
        self.description = description
        self.icon = icon
        self.name = name


class Reaction(DiscordObject):
    """Represents a Reaction.

    .. versionadded:: 0.2.0

    Attributes:
        count (:obj:`int`): times this emoji has been used to react
        me (:obj:`bool`): whether the current user reacted using this emoji
        emoji (:class:`.Emoji`): emoji information
    """

    def __init__(self, count=0, me=False, emoji=None):
        self.count = count
        self.me = me
        self.emoji = emoji

    async def _from_api_ext(self, key, value):
        if key == 'emoji':
            setattr(self, key, await Emoji.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)


class EmbedThumbnail(DiscordObject):
    """Represents a embed thumbnail object

    .. versionadded:: 0.2.0

    Attributes:
        url (:obj:`str`): source url of thumbnail (only supports http(s) and attachments)
        proxy_url (:obj:`str`): a proxied url of the thumbnail
        height (:obj:`int`): height of thumbnail
        width (:obj:`int`): width of thumbnail
    """

    def __init__(self, url="", proxy_url="", height=0, width=0):
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width


class EmbedVideo(DiscordObject):
    """Represents a embed video

    .. versionadded:: 0.2.0

    Attributes:
        url (:obj:`str`): source url of video
        height (:obj:`int`): height of video
        width (:obj:`int`): width of video
    """

    def __init__(self, url="", height=0, width=0):
        self.url = url
        self.height = height
        self.width = width


class EmbedImage(DiscordObject):
    """Represents a embed image

    .. versionadded:: 0.2.0

    Attributes:
        url (:obj:`str`): source url of image (only supports http(s) and attachments)
        proxy_url (:obj:`str`): a proxied url of the image
        height (:obj:`int`): height of image
        width (:obj:`int`): width of image
    """

    def __init__(self, url="", proxy_url="", height=0, width=0):
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width


class EmbedProvider(DiscordObject):
    """Represents a embed provider

    .. versionadded:: 0.2.0

    Attributes:
        name (:obj:`str`): name of provider
        url (:obj:`str`): url of provider
    """

    def __init__(self, name="", url=""):
        self.name = name
        self.url = url


class EmbedAuthor(DiscordObject):
    """Represents a embed author

    .. versionadded:: 0.2.0

    Attributes:
        name (:obj:`str`): name of author
        url (:obj:`str`): url of author
        icon_url (:obj:`str`): url of author icon (only supports http(s) and attachments)
        proxy_icon_url (:obj:`str`): a proxied url of author icon
    """

    def __init__(self, name="", url="", icon_url="", proxy_icon_url=""):
        self.name = name
        self.url = url
        self.icon_url = icon_url
        self.proxy_icon_url = proxy_icon_url


class EmbedFooter(DiscordObject):
    """Represents a embed footer

    .. versionadded:: 0.2.0

    Attributes:
        text (:obj:`str`): footer text
        icon_url (:obj:`str`): url of footer icon (only supports http(s) and attachments)
        proxy_icon_url (:obj:`str`): a proxied url of footer icon
    """

    def __init__(self, text="", icon_url="", proxy_icon_url=""):
        self.text = text
        self.icon_url = icon_url
        self.proxy_icon_url = proxy_icon_url


class EmbedField(DiscordObject):
    """Represents a embed field

    .. versionadded:: 0.2.0

    Attributes:
        name (:obj:`str`): name of the field
        value (:obj:`str`): value of the field
        inline (:obj:`bool`): whether or not this field should display inline
    """

    def __init__(self, name="", value="", inline=False):
        self.name = name
        self.value = value
        self.inline = inline


class Embed(DiscordObject):
    """Represents a discord Embed

    .. versionadded:: 0.2.0

    Attributes:
        title (:obj:`str`): title of embed
        type (:obj:`str`): type of embed (always "rich" for webhook embeds)
        description (:obj:`str`): description of embed
        url (:obj:`str`): url of embed
        timestamp (:obj:`int`): timestamp of embed content
        color (:obj:`int`): color code of the embed
        footer (:class:`.EmbedFooter`): footer information
        image (:class:`.EmbedImage`): image information
        thumbnail (:class:`.EmbedThumbnail`): thumbnail information
        video (:class:`.EmbedVideo`): video information
        provider (:class:`.EmbedProvider`): provider information
        author (:class:`.EmbedAuthor`): author information
        fields (:obj:`list` of :class:`.EmbedField`): fields information
    """

    def __init__(self, title="", type="", description="", url="", timestamp=None,
                 color=0, footer=EmbedFooter(), image=EmbedImage(), thumbnail=EmbedThumbnail(),
                 video=EmbedVideo(), provider=EmbedProvider(), author=EmbedAuthor(), fields=[]):
        self.title = title
        self.type = type
        self.description = description
        self.url = url
        self.timestamp = timestamp
        self.color = color
        self.footer = footer
        self.image = image
        self.thumbnail = thumbnail
        self.video = video
        self.provider = provider
        self.author = author
        self.fields = fields

    async def _from_api_ext(self, key, value):
        if key == 'footer':
            setattr(self, key, await EmbedFooter.from_api_res(value))
        elif key == 'image':
            setattr(self, key, await EmbedImage.from_api_res(value))
        elif key == 'thumbnail':
            setattr(self, key, await EmbedThumbnail.from_api_res(value))
        elif key == 'video':
            setattr(self, key, await EmbedVideo.from_api_res(value))
        elif key == 'provider':
            setattr(self, key, await EmbedProvider.from_api_res(value))
        elif key == 'author':
            setattr(self, key, await EmbedAuthor.from_api_res(value))
        elif key == 'fields':
            setattr(self, key, [await EmbedField.from_api_res(x) for x in value])
        else:
            return await super()._from_api_ext(key, value)


class Attachment(DiscordObject):
    """Represents a attachment

    .. versionadded:: 0.2.0

    Attributes:
        id (:obj:`int`): attachment id
        filename (:obj:`str`): name of file attached
        size (:obj:`int`): size of file in bytes
        url (:obj:`str`): source url of file
        proxy_url (:obj:`str`): a proxied url of file
        height (:obj:`int`): height of file (if image)
        width (:obj:`int`): width of file (if image)
    """

    def __init__(self, id=0, filename="", size=0, url="", proxy_url="",
                 height=0, width=0):
        self.id = id
        self.filename = filename
        self.size = size
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width


class ChannelMessage(DiscordObject):
    """Represents a message sent in a channel within Discord.

    .. versionadded:: 0.2.0

    Note:
        The author object follows the structure of the :class:`.User` object, but is only a valid user in the case where the message is generated by a user or bot user.
        If the message is generated by a :class:`.Webhook`, the author object corresponds to the webhook's id, username, and avatar.
        You can tell if a message is generated by a webhook by checking for the webhook_id on the message object.

    Attributes:
        id (:obj:`int`): id of the message
        channel_id (:obj:`int`): id of the channel the message was sent in
        author (user): object the author of this message (not guaranteed to be a valid user, see below)
        content (:obj:`str`): contents of the message
        timestamp (:obj:`int`): timestamp when this message was sent
        edited_timestamp (:obj:`int`): timestamp when this message was edited (or null if never)
        tts (:obj:`bool`): whether this was a TTS message
        mention_everyone (:obj:`bool`): whether this message mentions everyone
        mentions (:obj:`list` of :class:`.User`): objects users specifically mentioned in the message
        mention_roles (:obj:`list` of :class:`.Role`): object ids roles specifically mentioned in this message
        attachments (:obj:`list` of :class:`.Attachment`): objects any attached files
        embeds (:obj:`list` of :class:`.Embed`): objects any embedded content
        reactions (:obj:`list` of :class:`.Reaction`): objects reactions to the message
        nonce (:obj:`int`, optional): used for validating a message was sent
        pinned (:obj:`bool`): whether this message is pinned
        webhook_id (:obj:`int`, optional): if the message is generated by a webhook, this is the webhook's id
        type (:obj:`int`): type of message
        activity (MessageActivity): activity object sent with Rich Presence-related chat embeds
        application (MessageApplication): application object sent with Rich Presence-related chat embeds
    """

    def __init__(self, id=0, channel_id=0, author=None, content="", timestamp=None,
                 edited_timestamp=None, tts=False, mention_everyone=False, mentions=[],
                 mention_roles=[], attachments=[], embeds=[], reactions=[],
                 nonce=0, pinned=False, webhook_id=0, type=0,
                 activity=MessageActivity(), application=MessageApplication()):
        self.id = id
        self.channel_id = channel_id
        self.author = author
        self.content = content
        self.timestamp = timestamp
        self.edited_timestamp = edited_timestamp
        self.tts = tts
        self.mention_everyone = mention_everyone
        self.mentions = mentions
        self.mention_roles = mention_roles
        self.attachments = attachments
        self.embeds = embeds
        self.reactions = reactions
        self.nonce = nonce
        self.pinned = pinned
        self.webhook_id = webhook_id
        self.type = type
        self.activity = activity
        self.application = application

    async def _from_api_ext(self, key, value):
        if key == 'author':
            setattr(self, key, await User.from_api_res(value))
        elif key == 'mentions':
            setattr(self, key, [await User.from_api_res(x) for x in value])
        elif key == 'mention_roles':
            setattr(self, key, [await Role.from_api_res(x) for x in value])
        elif key == 'attachments':
            setattr(self, key, [await Attachment.from_api_res(x) for x in value])
        elif key == 'reactions':
            setattr(self, key, [await Reaction.from_api_res(x) for x in value])
        elif key == 'activity':
            setattr(self, key, await MessageActivity.from_api_res(value))
        elif key == 'activity':
            setattr(self, key, await MessageApplication.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)


__all__ = [
    'Channel',
    'ChannelMessage',
    'Overwrite',
    'MessageActivity',
    'MessageApplication',
    'Reaction',
    'Embed',
    'EmbedThumbnail',
    'EmbedVideo',
    'EmbedImage',
    'EmbedProvider',
    'EmbedAuthor',
    'EmbedFooter',
    'EmbedField',
    'Attachment',
]



from .base import DiscordObject


class Role(DiscordObject):
    """Represents a discord role

    .. versionadded:: 0.2.0

    Attributes:
        id (:obj:`int`): role id
        name (:obj:`str`): role name
        color (:obj:`int`): integer representation of hexadecimal color code
        hoist (:obj:`bool`): if this role is pinned in the user listing
        position (:obj:`int`): position of this role
        permissions (:obj:`int`): permission bit set
        managed (:obj:`bool`): whether this role is managed by an integration
        mentionable (:obj:`bool`): whether this role is mentionable
    """

    def __init__(self, id=0, name="", color=0, hoist=False, position=0,
                 permissions=0, managed=False, mentionable=False):
        self.id = id
        self.name = name
        self.color = color
        self.hoist = hoist
        self.position = position
        self.permissions = permissions
        self.managed = managed
        self.mentionable = mentionable

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'<Role Object: {self.name}#{self.id}>'


__all__ = [
    'Role',
]



from .base import DiscordObject


class ActivityParty(DiscordObject):
    """Activity party

    .. versionadded:: 0.2.0

    Attributes:
        id (:obj:`str`, optional): the id of the party
        size (:obj:`list` of :obj:`int`): array of two integers (current_size, max_size), used to show the party's current and maximum size
    """

    def __init__(self, id="", size=[]):
        self.id = id
        self.size = size


class ActivityTimestamps(DiscordObject):
    """Activity timestamps

    .. versionadded:: 0.2.0

    Attributes:
        start (:obj:`int`, optional): unix time (in milliseconds) of when the activity started
        end (:obj:`int`, optional): unix time (in milliseconds) of when the activity ends
    """

    def __init__(self, start=0, end=0):
        self.start = start
        self.end = end


class ActivityAssets(DiscordObject):
    """Activity assets

    .. versionadded:: 0.2.0

    Attributes:
        large_image (:obj:`str`, optional): the id for a large asset of the activity, usually a snowflake
        large_text (:obj:`str`, optional): text displayed when hovering over the large image of the activity
        small_image (:obj:`str`, optional): the id for a small asset of the activity, usually a snowflake
        small_text (:obj:`str`, optional): text displayed when hovering over the small image of the activity
    """

    def __init__(self, large_image="", large_text="", small_image="", small_text=""):
        self.large_image = large_image
        self.large_text = large_text
        self.small_image = small_image
        self.small_text = small_text


class Activity(DiscordObject):
    """Represents a discord activity

    .. versionadded:: 0.2.0

    Attributes:
        name (:obj:`str`): the activity's name
        type (:obj:`int`): activity type
        url (:obj:`str`, optional): stream url, is validated when type is 1
        timestamps (:class:`.Timestamps`): object unix timestamps for start and/or end of the game
        application_id (:obj:`int`, optional): application id for the game
        details (:obj:`str`, optional): what the player is currently doing
        state (:obj:`str`, optional): the user's current party status
        party (:class:`.Party`): object information for the current party of the player
        assets (:class:`.Assets`): object images for the presence and their hover texts
    """

    def __init__(self, name="", type=0, url="", timestamps=None, application_id=0,
                 details="", state="", party=None, assets=None):
        self.name = name
        self.type = type
        self.url = url
        self.timestamps = timestamps
        self.application_id = application_id
        self.details = details
        self.state = state
        self.party = party
        self.assets = assets

    async def _from_api_ext(self, key, value):
        if key == 'timestamps':
            setattr(self, key, await ActivityTimestamps.from_api_res(value))
        elif key == 'party':
            setattr(self, key, await ActivityParty.from_api_res(value))
        elif key == 'assets':
            setattr(self, key, await ActivityAssets.from_api_res(value))
        else:
            await super()._from_api_ext(key, value)


__all__ = [
    'Activity',
    'ActivityParty',
    'ActivityTimestamps',
    'ActivityAssets',
]



import sys


class WebSocketCreationError(Exception):
    def __init__(self):
        print("Error creating a websocket to connect to discord.", file=sys.stderr)


class EventTypeError(Exception):
    pass


class AuthorizationError(Exception):
    def __init__(self):
        print(
            'You requested a api endpoint which you have no authorization', file=sys.stderr)


class UnhandledEndpointStatusError(Exception):
    pass


__all__ = [
    'WebSocketCreationError',
    'EventTypeError',
    'AuthorizationError',
]



import asyncio
import aiohttp
import json
import platform

from .enums import GatewayOpcodes
from .guild import Guild, GuildMember
from .user import User
from .http import HTTPHandler
from .role import Role
from .channel import Channel, ChannelMessage
from .emoji import Emoji
from .activity import Activity
from .voice import VoiceState

import logging
logger = logging.getLogger(__name__)


class DiscordWebsocket:
    """Class used for handling the websocket connection with the discord gateway

    .. versionadded:: 0.2.0

    Attributes:
        heartbeat_interval (:obj:`float`): The interval to send pings
        _trace (:obj:`str`): Used for debugging
        seq (:obj:`str`): Used in pings
        session_id (:obj:`str`): Used for resuming
        http (:class:`HTTPHandler`): Used for sending http requests and session handling.
        gateway_url (:obj:`str`): The gateway url
        shards (:obj:`list` of :obj:`int`): Used for opening multiple connections
        ws (:class:aiohttp.ClientWebSocketResponse``): The websocket
    """
    def __init__(self, http: HTTPHandler=None, session_id: str=None, shards: list=[]):
        self.heartbeat_interval: float = None
        self._trace: list = []
        self.seq: str = None
        self.heartbeat_future: asyncio.Future = None
        self.session_id: str = session_id
        self.http: HTTPHandler = http
        self.gateway_url: str = None
        self.shards: list = shards
        self.ws: aiohttp.ClientWebSocketResponse = None

    @property
    def closed(self) -> bool:
        return self.ws is None or self.ws.closed

    async def send_heartbeat(self):
        while True:
            if self.ws is None or self.ws.closed:
                break
            logger.debug(
                f'Waiting {self.heartbeat_interval / 1000} seconds to send heartbeat')
            await asyncio.sleep(self.heartbeat_interval / 1000)
            await self.ws.send_json({
                'op': 1,
                'd': self.s
            })
            logger.debug("Sent heartbeat")

    async def close(self) -> bool:
        """Closes the websocket

        .. versionadded:: 0.2.0

        Returns:
            bool: True if succeded closing. False if the websocket was already closed
        """
        if self.ws is not None and not self.ws.closed:
            if self.heartbeat_future is not None and not self.heartbeat_future.cancelled():
                self.heartbeat_future.cancel()
            await self.ws.close()
            logger.debug('Websocket closed!')
            return True
        else:
            return False

    async def start(self):
        """Starts the websocket

        .. versionadded:: 0.2.0
        """
        if self.ws is not None and not self.ws.closed:
            await self.ws.close()
            self.ws = None

        info = await self.http.request_url('/gateway/bot')
        self.gateway_url = info['url']
        self.shards = info['shards']

        logger.debug(f'I can use {self.shards} shards!')

        async with self.http.session.ws_connect(self.gateway_url + '?v=6&encoding=json') as ws:
            self.ws = ws
            async for msg in self.ws:
                # logger.debug(msg)
                if msg.type == aiohttp.WSMsgType.TEXT:
                    dct = json.loads(msg.data)
                    opcode = dct['op']
                    data = dct['d']
                    if GatewayOpcodes(opcode) == GatewayOpcodes.HELLO:
                        if self.heartbeat_future is not None:
                            if not self.heartbeat_future.cancelled():
                                self.heartbeat_future.cancel()
                                self.heartbeat_future = None
                        self.heartbeat_interval = dct['d']['heartbeat_interval']
                        self._trace = dct['d']['_trace']
                        self.seq = dct['s']
                        if self.session_id is not None:
                            pass
                            await ws.send_json({
                                "op": 6,  # Resume
                                "d": {
                                    "token": self.http.token,
                                    "session_id": self.session_id,
                                    "seq": self.seq
                                }
                            })
                        else:
                            await ws.send_json({
                                "op": 2,  # Identify
                                "d": {
                                    "token": self.http.token,
                                    "properties": {
                                        '$os': platform.system(),
                                        '$browser': 'discord.aio',
                                        '$device': 'discord.aio'
                                    },
                                    "compress": False,
                                    "large_threshold": 250
                                }
                            })
                        logger.debug(
                            f'Ensuring to heartbeat every {self.heartbeat_interval / 1000} seconds!')
                        self.heartbeat_future = asyncio.ensure_future(
                            self.send_heartbeat())
                    elif GatewayOpcodes(opcode) == GatewayOpcodes.HEARTBEAT_ACK:
                        logger.debug(
                            f'Got HEARTBEAT_ACK (new s: {dct["s"]}, old s: {self.seq})')
                        self.s = dct['s']
                    elif GatewayOpcodes(opcode) == GatewayOpcodes.DISPATCH:
                        event_type = dct['t']
                        asyncio.ensure_future(
                            self.dispatch_event(event_type, data))

    async def dispatch_event(self, event, data):
        if event == 'READY':
            self.session_id = data['session_id']
            self.http.get_client().user = await User.from_api_res(data['user'])
            self.http.get_client().guilds = await Guild.from_api_res(data['guilds'])
            return await self.http.get_client().raise_event('on_ready')

        elif event == 'RESUMED':
            return await self.http.get_client().raise_event('on_resumed')

        elif event == 'INVALID_SESSION':
            return await self.http.get_client().raise_event('on_invalid_session', data)

        elif event == 'CHANNEL_CREATE':
            return await self.http.get_client().raise_event('on_channel_create', await Channel.from_api_res(data))

        elif event == 'CHANNEL_UPDATE':
            return await self.http.get_client().raise_event('on_channel_update', await Channel.from_api_res(data))

        elif event == 'CHANNEL_DELETE':
            return await self.http.get_client().raise_event('on_channel_delete', await Channel.from_api_res(data))

        elif event == 'CHANNEL_PINS_UPDATE':
            return await self.http.get_client().raise_event('on_channel_pin', data['channel_id'], data['last_pin_timestamp'])

        elif event == 'GUILD_CREATE':
            guild = await Guild.from_api_res(data)
            return await self.http.get_client().raise_event('on_guild_create', guild)

        elif event == 'GUILD_DELETE':
            guild = await Guild.from_api_res(data)
            return await self.http.get_client().raise_event('on_guild_delete', guild)

        elif event == 'GUILD_BAN_ADD':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_ban', guild_id, await User.from_api_res(data))

        elif event == 'GUILD_BAN_REMOVE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_ban_remove', guild_id, await User.from_api_res(data))

        elif event == 'GUILD_EMOJIS_UPDATE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_emojis_update', guild_id, await Emoji.from_api_res(data['emojis']))

        elif event == 'GUILD_INTEGRATIONS_UPDATE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_integrations_update', guild_id)

        elif event == 'GUILD_MEMBER_ADD':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_member_add', guild_id, await GuildMember.from_api_res(data))

        elif event == 'GUILD_MEMBER_REMOVE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_member_remove', guild_id, await User.from_api_res(data['user']))

        elif event == 'GUILD_MEMBER_UPDATE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_member_update', guild_id,
                                                            await Role.from_api_res(data['roles']), await User.from_api_res(data['user']), data['nick'])

        elif event == 'GUILD_MEMBERS_CHUNK':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_members_chunk', guild_id, await GuildMember.from_api_res(data['members']))

        elif event == 'GUILD_ROLE_CREATE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_role_create', guild_id, await Role.from_api_res(data['role']))

        elif event == 'GUILD_ROLE_UPDATE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_role_update', guild_id, await Role.from_api_res(data['role']))

        elif event == 'GUILD_ROLE_DELETE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_guild_role_delete', guild_id, data['role_id'])

        elif event == 'MESSAGE_CREATE':
            message = await ChannelMessage.from_api_res(data)
            await self.http.get_client().raise_event('on_message', message)

        elif event == 'MESSAGE_UPDATE':
            message = await ChannelMessage.from_api_res(data)
            await self.http.get_client().raise_event('on_message_create', message)

        elif event == 'MESSAGE_DELETE':
            await self.http.get_client().raise_event('on_message_delete', data['id'], data['channel_id'])

        elif event == 'MESSAGE_DELETE_BULK':
            await self.http.get_client().raise_event('on_message_delete_bulk', data['ids'], data['channel_id'])

        elif event == 'MESSAGE_REACTION_ADD':
            emoji = await Emoji.from_api_res(data['emoji'])
            await self.http.get_client().raise_event(
                'on_message_reaction_add', data['user_id'], data['channel_id'], data['message_id'], emoji)

        elif event == 'MESSAGE_REACTION_REMOVE':
            emoji = await Emoji.from_api_res(data['emoji'])
            await self.http.get_client().raise_event(
                'on_message_reaction_remove', data['user_id'], data['channel_id'], data['message_id'], emoji)

        elif event == 'MESSAGE_REACTION_REMOVE_ALL':
            await self.http.get_client().raise_event(
                'on_message_reaction_remove_all', data['channel_id'], data['message_id'])

        elif event == 'PRESENCE_UPDATE':
            await self.http.get_client().raise_event('on_presence_update', await User.from_api_res(data['user']),
                                                     data['roles'], await Activity.from_api_res(data.get('game')), data['guild_id'],
                                                     data['status'])

        elif event == 'TYPING_START':
            await self.http.get_client().raise_event('on_typing_start', data['user_id'], data['channel_id'], data['timestamp'])

        elif event == 'USER_UPDATE':
            await self.http.get_client().raise_event('on_user_update', await User.from_api_res(data))

        elif event == 'VOICE_STATE_UPDATE':
            await self.http.get_client().raise_event('on_voice_state_update', await VoiceState.from_api_res(data))

        elif event == 'VOICE_SERVER_UPDATE':
            await self.http.get_client().raise_event('on_voice_server_update', data['token'], data['guild_id'], data['endpoint'])

        elif event == 'WEBHOOKS_UPDATE':
            await self.http.get_client().raise_event('on_webhooks_update', data['guild_id'], data['channel_id'])

        else:
            logger.critical(
                f'Unhandled event type {event}, data: {data}')


__all__ = [
    'DiscordWebsocket',
]



"""
discord.aio is an asynchronous Discord API wrapper for python 3.6+
"""

from .client import DiscordBot
from .channel import Channel, ChannelMessage, Attachment, Embed, EmbedAuthor, EmbedField, EmbedFooter, EmbedImage, EmbedProvider, EmbedThumbnail, EmbedVideo
from .emoji import Emoji
from .user import User, UserConnection
from .guild import Guild, GuildEmbed, GuildMember, Integration, IntegrationAccount
from .role import Role
from .http import HTTPHandler
from .websocket import DiscordWebsocket
from .webhook import Webhook
from .invite import Invite, InviteMetadata
from .enums import ChannelTypes, ExplicitContentFilterLevel, MessageActivityTypes, MessageNotificationLevel, MFALevel, VerificationLevel
from .exceptions import WebSocketCreationError, AuthorizationError, EventTypeError, UnhandledEndpointStatusError
from .base import DiscordObject
from .voice import VoiceRegion, VoiceState
from .activity import Activity, ActivityAssets, ActivityParty, ActivityTimestamps
from .constants import DISCORD_API_URL, DISCORD_CDN
from .version import __version__

__author__ = 'Ryozuki'
__license__ = 'MIT'



"""Webhooks are a low-effort way to post messages to channels in Discord. They do not require a bot user or authentication to use."""

from .base import DiscordObject


class Webhook(DiscordObject):
    """Used to represent a webhook.

    .. versionadded:: 0.2.0

    Attributes:
        id (:obj:`int`): the id of the webhook
        guild_id (:obj:`int`, optional): the guild id this webhook is for
        channel_id (:obj:`int`): the channel id this webhook is for
        user (:class:`.User`): the user this webhook was created by (not returned when getting a webhook with its token)
        name (:obj:`str`): the default name of the webhook
        avatar (:obj:`str`): the default avatar of the webhook
        token (:obj:`str`): the secure token of the webhook
    """

    def __init__(self, id=0, guild_id=0, channel_id=0, user=None, name="",
                 avatar="", token=""):
        self.id = id
        self.guild_id = guild_id
        self.channel_id = channel_id
        self.user = user
        self.name = name
        self.avatar = avatar
        self.token = token


__all__ = [
    'Webhook',
]



from typing import Optional
from .base import DiscordObject


class VoiceState(DiscordObject):
    """Used to represent a user's voice connection status.
    
    Attributes:
        guild_id (:obj:`int`, optional): the guild id this voice state is for
        channel_id (:obj:`int`): the channel id this user is connected to
        user_id (:obj:`int`): the user id this voice state is for
        session_id (:obj:`str`): the session id for this voice state
        deaf (:obj:`bool`): whether this user is deafened by the server
        mute (:obj:`bool`): whether this user is muted by the server
        self_deaf (:obj:`bool`): whether this user is locally deafened
        self_mute (:obj:`bool`): whether this user is locally muted
        suppress (:obj:`bool`): whether this user is muted by the current user
    """
    def __init__(self, guild_id: Optional[int]=None, channel_id: int=0, user_id: int=0, session_id: str='', deaf=False,
                 mute=False, self_deaf=False, self_mute=False, suppress=False):
        self.guild_id: int = guild_id
        self.channel_id = channel_id
        self.user_id = user_id
        self.session_id = session_id
        self.deaf = deaf
        self.mute = mute
        self.self_deaf = self_deaf
        self.self_mute = self_mute
        self.suppress = suppress


class VoiceRegion(DiscordObject):
    """
    Attributes:
        id (:obj:`str`): unique ID for the region
        name (:obj:`str`): name of the region
        vip (:obj:`bool`): true if this is a vip-only server
        optimal (:obj:`bool`): true for a single server that is closest to the current user's client
        deprecated (:obj:`bool`): whether this is a deprecated voice region (avoid switching to these)
        custom (:obj:`bool`): whether this is a custom voice region (used for events/etc)
    """
    def __init__(self, id="", name="", vip=False, optimal=False, deprecated=False,
                 custom=False):
        self.id = id
        self.name = name
        self.vip = vip
        self.optimal = optimal
        self.deprecated = deprecated
        self.custom = custom


__all__ = [
    'VoiceState',
    'VoiceRegion',
]
